SystemOrganization addCategory: #ObjectUsageTracer!

!Behavior methodsFor: '*ObjectUsageTracer' stamp: 'MarianoMartinezPeck 12/23/2011 12:05'!
classesForlookupSelector: selector
	"Look up the given selector in my methodDictionary.
	Return the corresponding method if found.
	Otherwise chase the superclass chain and try again.
	Return nil if no method is found."
	| lookupClass classes |
	classes := OrderedCollection new.
	lookupClass := self.
	[lookupClass == nil]
		whileFalse: 
			[ classes add: lookupClass.
			(lookupClass includesSelector: selector)
				ifTrue: [^ classes].
			lookupClass := lookupClass superclass].
	^ nil! !

!Behavior methodsFor: '*ObjectUsageTracer' stamp: 'MarianoMartinezPeck 12/23/2011 12:40'!
superclassesForlookupSelector: selector
	"Look up the given selector in my methodDictionary.
	Return the corresponding method if found.
	Otherwise chase the superclass chain and try again.
	Return nil if no method is found."
	| lookupClass classes |
	classes := OrderedCollection new.
	lookupClass := self.
	[lookupClass == nil]
		whileFalse: 
			[ (lookupClass includesSelector: selector)
				ifTrue: [^ classes].
			lookupClass := lookupClass superclass.
			classes add: lookupClass.].
	^ nil! !

!Object methodsFor: '*ObjectUsageTracer' stamp: 'MarianoMartinezPeck 12/23/2011 11:39'!
foo: anObject
	anObject name! !

Object subclass: #ObjectUsageTracer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ObjectUsageTracer'!

!ObjectUsageTracer methodsFor: 'enable - disable' stamp: 'MarianoMartinezPeck 12/22/2011 12:16'!
disableTracing
	self primitiveSetHasToTrace: false.
! !

!ObjectUsageTracer methodsFor: 'enable - disable' stamp: 'MarianoMartinezPeck 12/22/2011 11:34'!
enableTracing
	self primitiveSetHasToTrace: true.
! !

!ObjectUsageTracer methodsFor: 'helpers' stamp: 'MarianoMartinezPeck 12/22/2011 11:39'!
getStadistics
	| anArray stadistics|
	anArray := self primitiveGetStadistics.
	stadistics := ObjectUsageTracerStatistics fromPrimitiveArray: anArray.
	^ stadistics ! !

!ObjectUsageTracer methodsFor: 'marking' stamp: 'MarianoMartinezPeck 12/22/2011 11:54'!
isMarked: anObject
	^ anObject isExperimentalBitOn! !

!ObjectUsageTracer methodsFor: 'marking' stamp: 'MarianoMartinezPeck 12/22/2011 11:53'!
markObject: anObject
	anObject experimentalBit: true.! !

!ObjectUsageTracer methodsFor: 'primitives' stamp: 'MarianoMartinezPeck 12/22/2011 11:32'!
primitiveGetStadistics 

	<primitive: 154>	"successful completion returns an Array with the stadistics numbers"
	^  self primitiveFailed 
! !

!ObjectUsageTracer methodsFor: 'primitives' stamp: 'MarianoMartinezPeck 12/22/2011 11:33'!
primitiveSetHasToTrace: aBoolean

	<primitive: 153>	"successful completion returns self"
	^  self primitiveFailed 
! !

!ObjectUsageTracer methodsFor: 'start - stop' stamp: 'MarianoMartinezPeck 12/22/2011 11:36'!
startTracing
	self unmarkAllObjects.
	self enableTracing.
	! !

!ObjectUsageTracer methodsFor: 'start - stop' stamp: 'MarianoMartinezPeck 12/22/2011 11:36'!
stopTracing
	self disableTracing.
	! !

!ObjectUsageTracer methodsFor: 'start - stop' stamp: 'MarianoMartinezPeck 12/22/2011 11:37'!
stopTracingAndGetStadistics
	self disableTracing.
	^ self getStadistics.
! !

!ObjectUsageTracer methodsFor: 'tracing' stamp: 'MarianoMartinezPeck 12/22/2011 12:41'!
trace: aBlock
	self startTracing.
	aBlock value.
	self stopTracing.! !

!ObjectUsageTracer methodsFor: 'helpers' stamp: 'MarianoMartinezPeck 12/22/2011 11:35'!
unmarkAllObjects
	ExperimentalBitHandler turnOffExperimentalBitOfAllObjects! !

!ObjectUsageTracer methodsFor: 'marking' stamp: 'MarianoMartinezPeck 12/22/2011 12:17'!
unmarkObject: anObject
	anObject experimentalBit: false.! !

Object subclass: #ObjectUsageTracerStatistics
	instanceVariableNames: 'unusedCount usedCount objectCount unusedPercentage usedPercentage usedMemory unusedMemory totalMemory percentageUsedMemory percentageUnusedMemory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ObjectUsageTracer'!

!ObjectUsageTracerStatistics class methodsFor: 'as yet unclassified' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
fromPrimitiveArray: anArray
	| usedCount unusedCount usedMemory unusedMemory objectsCount totalMemory |
	usedCount := (anArray at: 1).
	unusedCount := (anArray at: 2).
	usedMemory  := (anArray at: 3).
	unusedMemory  := (anArray at: 4).
	
	objectsCount := usedCount + unusedCount.
	totalMemory := unusedMemory + usedMemory.
	^ self new
		usedCount: usedCount;
		unusedCount: unusedCount;
		objectCount: objectsCount;
		unusedPercentage: ((unusedCount * 100) / objectsCount) asFloat;
		usedPercentage: ((usedCount * 100) / objectsCount) asFloat;
		usedMemory: usedMemory;
		unusedMemory: unusedMemory;
		totalMemory: totalMemory;
		percentageUnusedMemory: ((unusedMemory * 100) / totalMemory) asFloat;
		percentageUsedMemory: ((usedMemory * 100) / totalMemory) asFloat
		 ! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
objectCount
	^ objectCount! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
objectCount: anObject
	objectCount := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
percentageUnusedMemory
	^ percentageUnusedMemory! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
percentageUnusedMemory: anObject
	percentageUnusedMemory := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
percentageUsedMemory
	^ percentageUsedMemory! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
percentageUsedMemory: anObject
	percentageUsedMemory := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
printOn: aStream
	aStream 
		nextPutAll: 'Total amount of objects: ', self objectCount asString, String cr;
		nextPutAll: 'Amount of used objects: ', self usedCount asString, String cr;
		nextPutAll: 'Percentage of used objects: ', self usedPercentage asString, String cr;
		nextPutAll: 'Amount of unused objects: ', self unusedCount asString, String cr;
		nextPutAll: 'Percentage of unused objects: ', self unusedPercentage asString, String cr;
		nextPutAll: String cr;
		nextPutAll: 'Total bytes of memory used: ', self totalMemory  asString, String cr;
		nextPutAll: 'Bytes of memory for the used objects: ', self usedMemory asString, String cr;
		nextPutAll: 'Percentage of memory for the used objects: ', self percentageUsedMemory asString, String cr;
		nextPutAll: 'Bytes of memory for the unused objects: ', self unusedMemory asString, String cr;
		nextPutAll: 'Percentage of memory for the unused objects: ', self percentageUnusedMemory asString, String cr.
	! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
totalMemory
	^ totalMemory! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
totalMemory: anObject
	totalMemory := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedCount
	^ unusedCount! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedCount: anObject
	unusedCount := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedMemory
	^ unusedMemory! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedMemory: anObject
	unusedMemory := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedPercentage
	^ unusedPercentage! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
unusedPercentage: anObject
	unusedPercentage := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedCount
	^ usedCount! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedCount: anObject
	usedCount := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedMemory
	^ usedMemory ! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedMemory: anObject
	usedMemory := anObject! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedPercentage
	^ usedPercentage! !

!ObjectUsageTracerStatistics methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 12/22/2011 11:38'!
usedPercentage: anObject
	usedPercentage := anObject! !

TestCase subclass: #ObjectUsageTracerTestCase
	instanceVariableNames: 'tracer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ObjectUsageTracer'!

!ObjectUsageTracerTestCase methodsFor: 'running' stamp: 'MarianoMartinezPeck 12/23/2011 11:32'!
setUp
	tracer := ObjectUsageTracer new.
	tracer unmarkAllObjects.
	tracer disableTracing.
	Object flushCache.! !

!ObjectUsageTracerTestCase methodsFor: 'tests - basic' stamp: 'MarianoMartinezPeck 12/22/2011 12:26'!
testManualMarkingAndUnmarking
	|  obj1 obj2 obj3 obj4 |
	
	obj1 := 'anObject'.
	tracer markObject: obj1.
	obj2 := nil.
	tracer markObject: obj2.
	obj3 := Time now.
	tracer markObject: obj3.
	obj4 := 9999999999999999999999.
	tracer markObject: obj4.

	self assert: (tracer isMarked: obj1).
	self assert: (tracer isMarked: obj1).
	self assert: (tracer isMarked: obj1).
	self assert: (tracer isMarked: obj1).
	
	tracer unmarkObject: obj1.
	tracer unmarkObject: obj2.
	tracer unmarkObject: obj3.
	tracer unmarkObject: obj4.
	
	self deny: (tracer isMarked: obj1).
	self deny: (tracer isMarked: obj1).
	self deny: (tracer isMarked: obj1).
	self deny: (tracer isMarked: obj1).
	! !

!ObjectUsageTracerTestCase methodsFor: 'tests - tracing' stamp: 'MarianoMartinezPeck 12/23/2011 12:21'!
testTraceLookUpHierarchy
	"It tests that duing the method lookup, all the method dictionaries and classes of the lookup are marked. The resulted compiled method should be marked as well."
	|  obj1 classes |
	
	obj1 := 'anObject'.
	self deny: (tracer isMarked: obj1).
	
	tracer trace: [obj1 name ].
	
	self assert: (tracer isMarked: obj1).
	classes := (obj1 class classesForlookupSelector: #name).
	classes do: [:each | 
		self assert: (tracer isMarked: each).
		self assert: (tracer isMarked: each methodDict).	
		].
	self assert: (tracer isMarked:  (classes last compiledMethodAt: #name)).	! !

!ObjectUsageTracerTestCase methodsFor: 'tests - tracing' stamp: 'MarianoMartinezPeck 12/23/2011 12:47'!
testTraceLookUpHierarchyMultipleTimes
	"It tests that duing the method lookup, all the method dictionaries and classes of the lookup are marked. The resulted compiled method should be marked as well. In this tests, we execute the method several times, which means it will be found in the cache."
	|  obj1 classes |
	
	obj1 := 'anObject'.
	self deny: (tracer isMarked: obj1).
	
	obj1 name.
	obj1 name.
	obj1 name.
	obj1 name.
	
	tracer trace: [obj1 name ].
	
	self assert: (tracer isMarked: obj1).
	"Since the method was found in the cache, the methodDicts and the superclasses may not have been used."
	self assert: (tracer isMarked: obj1 class).	
	self assert: (tracer isMarked:  (obj1 class lookupSelector: #name)).	! !

!ObjectUsageTracerTestCase methodsFor: 'tests - tracing' stamp: 'MarianoMartinezPeck 12/22/2011 12:49'!
testTraceOnlyReceiver
	|  obj1 |
	
	obj1 := 'anObject'.
	self deny: (tracer isMarked: obj1).
	tracer trace: [obj1 name ].
	self assert: (tracer isMarked: obj1).! !

!ObjectUsageTracerTestCase methodsFor: 'tests - tracing' stamp: 'MarianoMartinezPeck 12/23/2011 11:39'!
testTraceOnlyReceiverMethodInsideMethod

	|  obj1 obj2 |
	
	obj1 := 'anObject'.
	obj2 := 'anotherObject'.
	self deny: (tracer isMarked: obj1).
	self deny: (tracer isMarked: obj2). 
	
	obj1 foo: obj2.
	obj1 foo: obj2.
	obj1 foo: obj2.
	obj1 foo: obj2.
	
	tracer trace:  [obj1 foo: obj2].
	
	self assert: (tracer isMarked: obj1).
	self assert: (tracer isMarked: obj2).! !

!ObjectUsageTracerTestCase methodsFor: 'tests - tracing' stamp: 'MarianoMartinezPeck 12/22/2011 12:51'!
testTraceOnlyReceiverMultipleTimes
	"In this case, since we previously send several times #name, it should have been jitted, and it should be marked anyway"
	|  obj1 |
	
	obj1 := 'anObject'.
	self deny: (tracer isMarked: obj1).
	
	obj1 name.
	obj1 name.
	obj1 name.
	obj1 name.
	
	tracer trace:  [obj1 name].
	
	self assert: (tracer isMarked: obj1).! !

!ObjectUsageTracerTestCase methodsFor: 'tests - basic' stamp: 'MarianoMartinezPeck 12/22/2011 11:52'!
testUnmarkAllObjects
	|  obj1 obj2 obj3 obj4 |
	
	obj1 := 'anObject'.
	tracer markObject: obj1.
	obj2 := 'anotherObject'.
	tracer markObject: obj2.
	obj3 := Time now.
	tracer markObject: obj3.
	obj4 := 9999999999999999999999.
	tracer markObject: obj4.

	tracer unmarkAllObjects.
	
	self deny: (tracer isMarked: obj1).
	self deny: (tracer isMarked: obj1).
	self deny: (tracer isMarked: obj1).
	self deny: (tracer isMarked: obj1).
	! !
